@@ -6,25 +6,49 @@
  *	License as published by the Free Software Foundation version 2.1
  *	of the License.
  *
- * Copyright (c) 2003-2008 Thomas Graf <tgraf@suug.ch>
+ * Copyright (c) 2003-2012 Thomas Graf <tgraf@suug.ch>
  */
 
 /**
  * @ingroup core
  * @defgroup utils Utilities
+ *
+ * Collection of helper functions
+ *
  * @{
+ *
+ * Header
+ * ------
+ * ~~~~{.c}
+ * #include <netlink/utils.h>
+ * ~~~~
  */
 
 #include <netlink-local.h>
 #include <netlink/netlink.h>
 #include <netlink/utils.h>
 #include <linux/socket.h>
+#include <stdlib.h> /* exit() */
 
 /**
- * Debug level
+ * Global variable indicating the desired level of debugging output.
+ *
+ * Level | Messages Printed
+ * ----- | ---------------------------------------------------------
+ *     0 | Debugging output disabled
+ *     1 | Warnings, important events and notifications
+ *     2 | More or less important debugging messages
+ *     3 | Repetitive events causing a flood of debugging messages
+ *     4 | Even less important messages
+ *
+ * If available, the variable will be initialized to the value of the
+ * environment variable `NLDBG`. The default value is 0 (disabled).
+ *
+ * For more information, see section @core_doc{_debugging, Debugging}.
  */
 int nl_debug = 0;
 
+/** @cond SKIP */
 struct nl_dump_params nl_debug_dp = {
 	.dp_type = NL_DUMP_DETAILS,
 };
@@ -84,9 +108,10 @@ int __nl_read_num_str_file(const char *path, int (*cb)(long, const char *))
 
 	return 0;
 }
+/** @endcond */
 
 /**
- * @name Unit Pretty-Printing
+ * @name Pretty Printing of Numbers
  * @{
  */
 
@@ -97,6 +122,7 @@ int __nl_read_num_str_file(const char *path, int (*cb)(long, const char *))
  *
  * Cancels down a byte counter until it reaches a reasonable
  * unit. The chosen unit is assigned to \a unit.
+ * This function assume 1024 bytes in one kilobyte
  * 
  * @return The cancelled down byte counter in the new unit.
  */
@@ -125,30 +151,57 @@ double nl_cancel_down_bytes(unsigned long long l, char **unit)
  * @arg	l		bit counter
  * @arg unit		destination unit pointer
  *
- * Cancels downa bit counter until it reaches a reasonable
+ * Cancels down bit counter until it reaches a reasonable
  * unit. The chosen unit is assigned to \a unit.
+ * This function assume 1000 bits in one kilobit
  *
  * @return The cancelled down bit counter in the new unit.
  */
 double nl_cancel_down_bits(unsigned long long l, char **unit)
 {
-	if (l >= 1099511627776ULL) {
+	if (l >= 1000000000000ULL) {
 		*unit = "Tbit";
-		return ((double) l) / 1099511627776ULL;
-	} else if (l >= 1073741824) {
-		*unit = "Gbit";
-		return ((double) l) / 1073741824;
-	} else if (l >= 1048576) {
-		*unit = "Mbit";
-		return ((double) l) / 1048576;
-	} else if (l >= 1024) {
-		*unit = "Kbit";
-		return ((double) l) / 1024;
-	} else {
-		*unit = "bit";
-		return (double) l;
+		return ((double) l) / 1000000000000ULL;
 	}
-		
+
+	if (l >= 1000000000) {
+		*unit = "Gbit";
+		return ((double) l) / 1000000000;
+	}
+
+	if (l >= 1000000) {
+		*unit = "Mbit";
+		return ((double) l) / 1000000;
+	}
+
+	if (l >= 1000) {
+		*unit = "Kbit";
+		return ((double) l) / 1000;
+	}
+
+	*unit = "bit";
+	return (double) l;
+}
+
+int nl_rate2str(unsigned long long rate, int type, char *buf, size_t len)
+{
+	char *unit;
+	double frac;
+
+	switch (type) {
+	case NL_BYTE_RATE:
+		frac = nl_cancel_down_bytes(rate, &unit);
+		break;
+	
+	case NL_BIT_RATE:
+		frac = nl_cancel_down_bits(rate, &unit);
+		break;
+	
+	default:
+		BUG();
+	}
+
+	return snprintf(buf, len, "%.2f%s/s", frac, unit);
 }
 
 /**
@@ -193,6 +246,9 @@ double nl_cancel_down_us(uint32_t l, char **unit)
  *  - b,kb/k,m/mb,gb/g for bytes
  *  - bit,kbit/mbit/gbit
  *
+ * This function assume 1000 bits in one kilobit and
+ * 1024 bytes in one kilobyte
+ *
  * @return The number of bytes or -1 if the string is unparseable
  */
 long nl_size2int(const char *str)
@@ -208,13 +264,13 @@ long nl_size2int(const char *str)
 		else if (!strcasecmp(p, "gb") || !strcasecmp(p, "g"))
 			l *= 1024*1024*1024;
 		else if (!strcasecmp(p, "gbit"))
-			l *= 1024*1024*1024/8;
+			l *= 1000000000L/8;
 		else if (!strcasecmp(p, "mb") || !strcasecmp(p, "m"))
 			l *= 1024*1024;
 		else if (!strcasecmp(p, "mbit"))
-			l *= 1024*1024/8;
+			l *= 1000000/8;
 		else if (!strcasecmp(p, "kbit"))
-			l *= 1024/8;
+			l *= 1000/8;
 		else if (!strcasecmp(p, "bit"))
 			l /= 8;
 		else if (strcasecmp(p, "b") != 0)
@@ -224,6 +280,56 @@ long nl_size2int(const char *str)
 	return l;
 }
 
+static const struct {
+	double limit;
+	const char *unit;
+} size_units[] = {
+	{ 1024. * 1024. * 1024. * 1024. * 1024., "EiB" },
+	{ 1024. * 1024. * 1024. * 1024., "TiB" },
+	{ 1024. * 1024. * 1024., "GiB" },
+	{ 1024. * 1024., "MiB" },
+	{ 1024., "KiB" },
+	{ 0., "B" },
+};
+
+/**
+ * Convert a size toa character string
+ * @arg size		Size in number of bytes
+ * @arg buf		Buffer to write character string to
+ * @arg len		Size of buf
+ *
+ * This function converts a value in bytes to a human readable representation
+ * of it. The function uses IEC prefixes:
+ *
+ * @code
+ * 1024 bytes => 1 KiB
+ * 1048576 bytes => 1 MiB
+ * @endcode
+ *
+ * The highest prefix is used which ensures a result of >= 1.0, the result
+ * is provided as floating point number with a maximum precision of 2 digits:
+ * @code
+ * 965176 bytes => 942.55 KiB
+ * @endcode
+ *
+ * @return pointer to buf
+ */
+char *nl_size2str(const size_t size, char *buf, const size_t len)
+{
+	size_t i;
+
+	for (i = 0; i < ARRAY_SIZE(size_units); i++) {
+		if (size >= size_units[i].limit) {
+			snprintf(buf, len, "%.2g%s",
+				(double) size / size_units[i].limit,
+				size_units[i].unit);
+			return buf;
+		}
+	}
+
+	BUG();
+}
+
 /**
  * Convert a character string to a probability
  * @arg str		probability encoded as character string
@@ -264,12 +370,13 @@ long nl_prob2int(const char *str)
  * @{
  */
 
-#ifdef USER_HZ
-static uint32_t user_hz = USER_HZ;
-#else
-static uint32_t user_hz = 100;
+#ifndef USER_HZ
+#define USER_HZ 100
 #endif
 
+static uint32_t user_hz = USER_HZ;
+static uint32_t psched_hz = USER_HZ;
+
 static double ticks_per_usec = 1.0f;
 
 /* Retrieves the configured HZ and ticks/us value in the kernel.
@@ -299,6 +406,8 @@ static void __init get_psched_settings(void)
 	if (!got_hz)
 		user_hz = sysconf(_SC_CLK_TCK);
 
+	psched_hz = user_hz;
+
 	if (getenv("TICKS_PER_USEC")) {
 		double t = strtod(getenv("TICKS_PER_USEC"), NULL);
 		ticks_per_usec = t;
@@ -313,11 +422,22 @@ static void __init get_psched_settings(void)
 			strncpy(name, "/proc/net/psched", sizeof(name) - 1);
 		
 		if ((fd = fopen(name, "r"))) {
-			uint32_t tick, us;
-			/* the file contains 4 hexadecimals, but we just use
-			   the first two of them */
-			fscanf(fd, "%08x %08x", &tick, &us);
-			ticks_per_usec = (double)tick/(double)us;
+			unsigned int ns_per_usec, ns_per_tick, nom, denom;
+
+			if (fscanf(fd, "%08x %08x %08x %08x",
+			       &ns_per_usec, &ns_per_tick, &nom, &denom) != 4) {
+                            fprintf(stderr, "Fatal error: can not read psched settings from \"%s\". " \
+                                    "Try to set TICKS_PER_USEC, PROC_NET_PSCHED or PROC_ROOT " \
+                                    "environment variables\n", name);
+                            exit(1);
+                        }
+
+			ticks_per_usec = (double) ns_per_usec / 
+					 (double) ns_per_tick;
+
+			if (nom == 1000000)
+				psched_hz = denom;
+
 			fclose(fd);
 		}
 	}
@@ -327,11 +447,18 @@ static void __init get_psched_settings(void)
 /**
  * Return the value of HZ
  */
-int nl_get_hz(void)
+int nl_get_user_hz(void)
 {
 	return user_hz;
 }
 
+/**
+ * Return the value of packet scheduler HZ
+ */
+int nl_get_psched_hz(void)
+{
+	return psched_hz;
+}
 
 /**
  * Convert micro seconds to ticks
@@ -404,10 +531,12 @@ int nl_str2msec(const char *str, uint64_t *result)
  */
 char * nl_msec2str(uint64_t msec, char *buf, size_t len)
 {
-	int i, split[5];
-	char *units[] = {"d", "h", "m", "s", "msec"};
+	uint64_t split[5];
+	size_t i;
+	static const char *units[5] = {"d", "h", "m", "s", "msec"};
+	char * const buf_orig = buf;
 
-#define _SPLIT(idx, unit) if ((split[idx] = msec / unit) > 0) msec %= unit
+#define _SPLIT(idx, unit) if ((split[idx] = msec / unit)) msec %= unit
 	_SPLIT(0, 86400000);	/* days */
 	_SPLIT(1, 3600000);	/* hours */
 	_SPLIT(2, 60000);	/* minutes */
@@ -415,18 +544,17 @@ char * nl_msec2str(uint64_t msec, char *buf, size_t len)
 #undef  _SPLIT
 	split[4] = msec;
 
-	memset(buf, 0, len);
-
-	for (i = 0; i < ARRAY_SIZE(split); i++) {
-		if (split[i] > 0) {
-			char t[64];
-			snprintf(t, sizeof(t), "%s%d%s",
-				 strlen(buf) ? " " : "", split[i], units[i]);
-			strncat(buf, t, len - strlen(buf) - 1);
-		}
+	for (i = 0; i < ARRAY_SIZE(split) && len; i++) {
+		int l;
+		if (split[i] == 0)
+			continue;
+		l = snprintf(buf, len, "%s%" PRIu64 "%s",
+			(buf==buf_orig) ? "" : " ", split[i], units[i]);
+		buf += l;
+		len -= l;
 	}
 
-	return buf;
+	return buf_orig;
 }
 
 /** @} */
@@ -436,7 +564,7 @@ char * nl_msec2str(uint64_t msec, char *buf, size_t len)
  * @{
  */
 
-static struct trans_tbl nlfamilies[] = {
+static const struct trans_tbl nlfamilies[] = {
 	__ADD(NETLINK_ROUTE,route)
 	__ADD(NETLINK_USERSOCK,usersock)
 	__ADD(NETLINK_FIREWALL,firewall)
@@ -477,7 +605,7 @@ int nl_str2nlfamily(const char *name)
  * @{
  */
 
-static struct trans_tbl llprotos[] = {
+static const struct trans_tbl llprotos[] = {
 	{0, "generic"},
 	__ADD(ARPHRD_ETHER,ether)
 	__ADD(ARPHRD_EETHER,eether)
@@ -506,6 +634,7 @@ static struct trans_tbl llprotos[] = {
 #ifdef ARPHRD_HWX25
 	__ADD(ARPHRD_HWX25,hwx25)
 #endif
+	__ADD(ARPHRD_CAN,can)
 	__ADD(ARPHRD_PPP,ppp)
 	__ADD(ARPHRD_HDLC,hdlc)
 	__ADD(ARPHRD_LAPB,lapb)
@@ -545,12 +674,17 @@ static struct trans_tbl llprotos[] = {
 	__ADD(ARPHRD_FCFABRIC+12,fcfb_12)
 	__ADD(ARPHRD_IEEE802_TR,tr)
 	__ADD(ARPHRD_IEEE80211,ieee802.11)
+	__ADD(ARPHRD_PHONET,phonet)
+	__ADD(ARPHRD_CAIF, caif)
 #ifdef ARPHRD_IEEE80211_PRISM
 	__ADD(ARPHRD_IEEE80211_PRISM, ieee802.11_prism)
 #endif
 #ifdef ARPHRD_VOID
 	__ADD(ARPHRD_VOID,void)
 #endif
+#ifdef ARPHRD_NONE
+	__ADD(ARPHRD_NONE,nohdr)
+#endif
 };
 
 char * nl_llproto2str(int llproto, char *buf, size_t len)
@@ -571,7 +705,7 @@ int nl_str2llproto(const char *name)
  * @{
  */
 
-static struct trans_tbl ether_protos[] = {
+static const struct trans_tbl ether_protos[] = {
 	__ADD(ETH_P_LOOP,loop)
 	__ADD(ETH_P_PUP,pup)
 	__ADD(ETH_P_PUPAT,pupat)
@@ -589,6 +723,7 @@ static struct trans_tbl ether_protos[] = {
 	__ADD(ETH_P_DIAG,diag)
 	__ADD(ETH_P_CUST,cust)
 	__ADD(ETH_P_SCA,sca)
+	__ADD(ETH_P_TEB,teb)
 	__ADD(ETH_P_RARP,rarp)
 	__ADD(ETH_P_ATALK,atalk)
 	__ADD(ETH_P_AARP,aarp)
@@ -597,6 +732,8 @@ static struct trans_tbl ether_protos[] = {
 #endif
 	__ADD(ETH_P_IPX,ipx)
 	__ADD(ETH_P_IPV6,ipv6)
+	__ADD(ETH_P_PAUSE,pause)
+	__ADD(ETH_P_SLOW,slow)
 #ifdef ETH_P_WCCP
 	__ADD(ETH_P_WCCP,wccp)
 #endif
@@ -605,7 +742,15 @@ static struct trans_tbl ether_protos[] = {
 	__ADD(ETH_P_MPLS_UC,mpls_uc)
 	__ADD(ETH_P_MPLS_MC,mpls_mc)
 	__ADD(ETH_P_ATMMPOA,atmmpoa)
+	__ADD(ETH_P_LINK_CTL,link_ctl)
 	__ADD(ETH_P_ATMFATE,atmfate)
+	__ADD(ETH_P_PAE,pae)
+	__ADD(ETH_P_AOE,aoe)
+	__ADD(ETH_P_TIPC,tipc)
+	__ADD(ETH_P_1588,ieee1588)
+	__ADD(ETH_P_FCOE,fcoe)
+	__ADD(ETH_P_FIP,fip)
+	__ADD(ETH_P_EDSA,edsa)
 	__ADD(ETH_P_EDP2,edp2)
 	__ADD(ETH_P_802_3,802.3)
 	__ADD(ETH_P_AX25,ax25)
@@ -616,6 +761,7 @@ static struct trans_tbl ether_protos[] = {
 	__ADD(ETH_P_WAN_PPP,wan_ppp)
 	__ADD(ETH_P_PPP_MP,ppp_mp)
 	__ADD(ETH_P_LOCALTALK,localtalk)
+	__ADD(ETH_P_CAN,can)
 	__ADD(ETH_P_PPPTALK,ppptalk)
 	__ADD(ETH_P_TR_802_2,tr_802.2)
 	__ADD(ETH_P_MOBITEX,mobitex)
@@ -623,6 +769,12 @@ static struct trans_tbl ether_protos[] = {
 	__ADD(ETH_P_IRDA,irda)
 	__ADD(ETH_P_ECONET,econet)
 	__ADD(ETH_P_HDLC,hdlc)
+	__ADD(ETH_P_ARCNET,arcnet)
+	__ADD(ETH_P_DSA,dsa)
+	__ADD(ETH_P_TRAILER,trailer)
+	__ADD(ETH_P_PHONET,phonet)
+	__ADD(ETH_P_IEEE802154,ieee802154)
+	__ADD(ETH_P_CAIF,caif)
 };
 
 char *nl_ether_proto2str(int eproto, char *buf, size_t len)
@@ -701,7 +853,7 @@ void nl_new_line(struct nl_dump_params *params)
 			else if (params->dp_buf)
 				strncat(params->dp_buf, " ",
 					params->dp_buflen -
-					sizeof(params->dp_buf) - 1);
+					strlen(params->dp_buf) - 1);
 		}
 	}
 
@@ -716,13 +868,15 @@ static void dump_one(struct nl_dump_params *parms, const char *fmt,
 		vfprintf(parms->dp_fd, fmt, args);
 	else if (parms->dp_buf || parms->dp_cb) {
 		char *buf = NULL;
-		vasprintf(&buf, fmt, args);
-		if (parms->dp_cb)
-			parms->dp_cb(parms, buf);
-		else
-			strncat(parms->dp_buf, buf,
-			        parms->dp_buflen - strlen(parms->dp_buf) - 1);
-		free(buf);
+		if (vasprintf(&buf, fmt, args) >= 0) {
+			if (parms->dp_cb)
+				parms->dp_cb(parms, buf);
+			else
+				strncat(parms->dp_buf, buf,
+					parms->dp_buflen -
+					strlen(parms->dp_buf) - 1);
+			free(buf);
+		}
 	}
 }
 
@@ -785,12 +939,14 @@ void __trans_list_clear(struct nl_list_head *head)
 		free(tl->a);
 		free(tl);
 	}
+
+	nl_init_list_head(head);
 }
 
-char *__type2str(int type, char *buf, size_t len, struct trans_tbl *tbl,
-		 size_t tbl_len)
+char *__type2str(int type, char *buf, size_t len,
+		 const struct trans_tbl *tbl, size_t tbl_len)
 {
-	int i;
+	size_t i;
 	for (i = 0; i < tbl_len; i++) {
 		if (tbl[i].i == type) {
 			snprintf(buf, len, "%s", tbl[i].a);
@@ -819,9 +975,9 @@ char *__list_type2str(int type, char *buf, size_t len,
 }
 
 char *__flags2str(int flags, char *buf, size_t len,
-		  struct trans_tbl *tbl, size_t tbl_len)
+		  const struct trans_tbl *tbl, size_t tbl_len)
 {
-	int i;
+	size_t i;
 	int tmp = flags;
 
 	memset(buf, 0, len);
@@ -838,11 +994,11 @@ char *__flags2str(int flags, char *buf, size_t len,
 	return buf;
 }
 
-int __str2type(const char *buf, struct trans_tbl *tbl, size_t tbl_len)
+int __str2type(const char *buf, const struct trans_tbl *tbl, size_t tbl_len)
 {
 	unsigned long l;
 	char *end;
-	int i;
+	size_t i;
 
 	if (*buf == '\0')
 		return -NLE_INVAL;
@@ -879,9 +1035,11 @@ int __list_str2type(const char *buf, struct nl_list_head *head)
 	return (int) l;
 }
 
-int __str2flags(const char *buf, struct trans_tbl *tbl, size_t tbl_len)
+int __str2flags(const char *buf, const struct trans_tbl *tbl, size_t tbl_len)
 {
-	int i, flags = 0, len;
+	int flags = 0;
+	size_t i;
+	size_t len; /* ptrdiff_t ? */
 	char *p = (char *) buf, *t;
 
 	for (;;) {
@@ -891,7 +1049,8 @@ int __str2flags(const char *buf, struct trans_tbl *tbl, size_t tbl_len)
 		t = strchr(p, ',');
 		len = t ? t - p : strlen(p);
 		for (i = 0; i < tbl_len; i++)
-			if (!strncasecmp(tbl[i].a, p, len))
+			if (len == strlen(tbl[i].a) &&
+			    !strncasecmp(tbl[i].a, p, len))
 				flags |= tbl[i].i;
 
 		if (!t)
